//--------------------------------------------------------------------------------
//   Simple list structure with free list 
//     Reference from yaneSDK4D
//     Copyright (c) kei mesuda 2007 
//     (http://www.yomogi.sakura.ne.jp/~si)
//--------------------------------------------------------------------------------
//debug;




class list(T)
{
//--------------------------------------------------
// List element class
//--------------------------------------------------
private:
    struct elem
    {
    public:
        elem*   pPrev;
        elem*   pNext;
        T       x;
    }




//--------------------------------------------------
// iterator class
//--------------------------------------------------
public:
    static class iterator
    {
    private:
        elem*   pel;
        
    public:
        this()            {}
        this(elem* pel_)  {pel = pel_;}

        // "it()" returns contents
        T opCall()
        {
            return pel.x;
        }

        // ==
        int opEquals(iterator it)
        {
            return cast(int)(pel == it.pel);
        }
        
        // it++
        iterator opPostInc()
        {
            iterator it = new iterator(pel); 
            inc();
            return it;
        }        

        // it--
        iterator opPostDec()
        {
            iterator it = new iterator(pel);
            dec();
            return it;
        }

        // ++it, it+=n
        iterator opAddAssign(int n)
        {
            for(int i=0; i<n; ++i) inc();
            return this;
        }

        // --it, it-=n
        iterator opSubAssign(int n)
        {
            for(int i=0; i<n; ++i) dec();
            return this;
        }

        elem* _pointer()
        {
            return pel;
        }
        
    // pointer operation
    private:
        void inc() { if (pel) pel = pel.pNext; }
        void dec() { if (pel) pel = pel.pPrev; }
    }




//--------------------------------------------------
// Member values
//--------------------------------------------------
private:
    // Terminal element
    // (term.prev = last_element; term.next = first_element)
    elem    term;    

    // Free list terminal element
    // (pool.prev = last_freelist_element; pool.next = (not use))
    elem    pool;

    // List member count
    int     count;
    int     alloc_count;

    // Terminal iterator
    iterator    it_term;


//--------------------------------------------------
// internal functions
//--------------------------------------------------
private:
    // Insert element
    void insert_elem(T x, elem* pelPrev, elem* pelNext)
    {
        elem* pelNew = alloc_elem(x);

        pelNew.pPrev  = pelPrev;
        pelNew.pNext  = pelNext;
        pelPrev.pNext = pelNew;
        pelNext.pPrev = pelNew;

        count++;
    }


    // Remove element
    T remove_elem(elem* pel)
    {
        // Strange if count == 0 but delel != term.
        assert(!(count == 0 && pel != &term));
        
        if (count == 0) return term.x;
        
        T ret = pel.x;

        pel.pPrev.pNext = pel.pNext;
        pel.pNext.pPrev = pel.pPrev;
        dealloc_elem(pel);
        
        count--;
        
        return ret;
    }
    
    
    // Allocate element
    elem* alloc_elem(T x)
    {
        elem* pelNew;
        
        if (pool.pPrev == &pool) {
            pelNew = new elem;
        } else {
            pelNew     = pool.pPrev;
            pool.pPrev = pelNew.pPrev;
        }
        
        pelNew.x = x;
        return pelNew;
    }

    
    // Deallocate element
    void dealloc_elem(elem* pel)
    {
        // initialize
        pel.x = T.init;
        
        // add to free list
        pel.pPrev  = pool.pPrev;
        pool.pPrev = pel;
    }

    
    // Deallocate all
    void dealloc_all()
    {
        elem* pel, pelPrev;
        for (pel=term.pPrev; pel!=&term; pel=pelPrev) {
            pelPrev = pel.pPrev;
            dealloc_elem(pel);
        }
    }
    
    
    // Delete free list
    void free_freelist()
    {
        elem* pel, pelPrev;
        for (pel=pool.pPrev; pel!=&pool; pel=pelPrev) {
            pelPrev = pel.pPrev;
            delete(pel);
        }
    }


    // Delete all
    void free_all()
    {
        for (elem* pel=term.pPrev; pel!=&term; pel=pel.pPrev) {
            pel.pPrev  = pool.pPrev;
            pool.pPrev = pel;
        }
        free_freelist();
    }




//--------------------------------------------------
//  Constructor
//--------------------------------------------------
public:
    this()
    {
        it_term = new iterator(&term);
        term.pNext = &term;
        term.pPrev = &term;
        pool.pPrev = &pool;
        count = 0;
        clear();
    }


    this(list lst)
    {
        this();
        foreach(T x; lst) push_back(x);
    }




//--------------------------------------------------
//  Reference
//--------------------------------------------------
public:
    // get size
    int size()
    {
        return count;
    }


    // same as size()
    int length()
    {
        return count;
    }


    // Is empty ?
    bool empty()
    {
        return (count == 0);
    }


    // get front
    T front()
    {
        return term.pNext.x;
    }


    // get back
    T back()
    {
        return term.pPrev.x;
    }




//--------------------------------------------------
//  Operation
//--------------------------------------------------
public:
    // "clear" delete all.
    void clear()
    {
        free_all();
        term.pNext = &term;
        term.pPrev = &term;
        pool.pPrev = &pool;
        count = 0;
    }

    
    // "erase" do not delete. stack to free-list
    void erase()
    {
        dealloc_all();
        term.pNext = &term;
        term.pPrev = &term;
        count = 0;
    }
    

    // remove
    T remove(iterator it)
    {
        return remove_elem(it._pointer());
    }
    
    
    // push front
    T push_front(T x) 
    {
        insert_elem(x, &term, term.pNext);
        return x;
    }


    // push back
    T push_back(T x)
    {
        insert_elem(x, term.pPrev, &term);
        return x;
    }


    // pop front
    T pop_front()
    {
        return remove_elem(term.pNext);
    }


    // pop back
    T pop_back()
    {
        return remove_elem(term.pPrev);
    }


    // pop front
    bool pop_front(out T x)
    {
        if (count == 0) return false;
        x = remove_elem(term.pNext);
        return true;
    }


    // pop back
    bool pop_back(out T x)
    {
        if (count == 0) return false;
        x = remove_elem(term.pPrev);
        return true;
    }



//--------------------------------------------------
//  Operation with iterator
//--------------------------------------------------
public:
    // begin iterator
    iterator begin()
    {
        return new iterator(term.pNext);
    }

    // rbegin iterator
    iterator rbegin()
    {
        return new iterator(term.pPrev);
    }

    // end
    iterator end()
    {
        return it_term;
    }

    // rend
    iterator rend()
    {
        return it_term;
    }
    
    
    
    
//--------------------------------------------------
//  Operator
//--------------------------------------------------
public:
    // Implement "~"
    list opCat(T x)
    {
        list newlist = new list(this);
        newlist.push_back(x);
        return newlist;
    }


    // Implement "~"
    list opCat(list lst)
    {
        list newlist = new list(this);
        foreach(T x; lst) newlist.push_back(x);
        return newlist;
    }


    // Implement "~="
    list opCatAssign(T x)
    {
        push_back(x);
        return this;
    }


    // Implement "~="
    list opCatAssign(list lst)
    {
        foreach(T x; lst) push_back(x);
        return this;
    }




//--------------------------------------------------
//  Operator foreach
//--------------------------------------------------
public:
    // Implement for foreach
    int opApply(int delegate(inout T) dg)
    {
        int   result;
        
        elem* pel = term.pNext; 
        while (pel != &term) 
        {
            result = dg(pel.x);
            if (result != 0) break;
            pel = pel.pNext;
        }
        
        return result;
    }


    // Implement for foreach with iterator
    int opApply(int delegate(inout iterator) dg)
    {
        int      result;
        iterator it;
        
        for (it = begin(); it != end(); ++it)
        {
            result = dg(it);
            if (result != 0) break;
        }
        
        return result;
    }


    // Implement for foreach with index
    int opApply(int delegate(inout int, inout T) dg)
    {
        int   result, idx;
        elem* pel;
        
        for (pel =  term.pNext, idx = 0; 
             pel != &term; 
             pel =  pel.pNext,  ++idx) 
        {
            result = dg(idx, pel.x);
            if (result != 0) break;
        }
        
        return result;
    }


    // Implement for foreach with iterator and index
    int opApply(int delegate(inout int, inout iterator) dg)
    {
        int      result, idx;
        iterator it;
        
        for (it =  begin(), idx = 0; 
             it != end(); 
             ++it,          ++idx)
        {
            result = dg(idx, it);
            if (result != 0) break;
        }
        
        return result;
    }


    // Implement for foreach_reverse 
    int opApplyReverse(int delegate(inout T) dg)
    {
        int   result;
        elem* pel;

        for (pel =  term.pPrev; 
             pel != &term; 
             pel =  pel.pPrev) 
        {
            result = dg(pel.x);
            if (result != 0) break;
        }
        
        return result;
    }


    // Implement for foreach_reverse with iterator
    int opApplyReverse(int delegate(inout iterator) dg)
    {
        int      result;
        iterator it;
        
        for (it = rbegin(); it != rend(); --it)
        {
            result = dg(it);
            if (result != 0) break;
        }
        
        return result;
    }


    // Implement for foreach_reverse with index
    int opApplyReverse(int delegate(inout int, inout T) dg)
    {
        int   result, idx;
        elem* pel;

        for (pel =  term.pPrev, idx = count - 1; 
             pel != &term; 
             pel =  pel.pPrev,  --idx) 
        {
            result = dg(idx, pel.x);
            if (result != 0) break;
        }
        
        return result;
    }


    // Implement for foreach_reverse with iterator and index
    int opApplyReverse(int delegate(inout int, inout iterator) dg)
    {
        int      result, idx;
        iterator it;
        
        for (it =  rbegin(), idx = count - 1;
             it != rend(); 
             --it,           --idx)
        {
            result = dg(idx, it);
            if (result != 0) break;
        }
        
        return result;
    }
}

