//------------------------------------------------------------------------------------------
// OpenGL Solid Model Class
//------------------------------------------------------------------------------------------
module util_opengl.model;

private import std.stdio;

private import opengl;
private import openglu;
private import std.file;
private import std.string;
private import std.regexp;
private import std.math;
private import std.path;
private import util.math;
private import util_opengl.texture;
private import util_opengl.material;
private import util_opengl.array;

private import vl = util.vector;




class CSolidModel
{
// member sturcture
//--------------------------------------------------
public:
    // face data structure
    static struct FACE {
    public:
        byte        mat;                // material index
        ushort      idx[3];             // vertex index
        vl.vector2  uv[3];              // texture coordination
        vl.vector3  normalFace;         // normal vector for face (calculate after loading)
        vl.vector3  normalVertex[3];    // vertex normal vector (calculate after loading)
    };


    // material data class
    static class ModelMaterial : CMaterial {
    public:
        char[]      name;               // material name
        bool        isVertexColor;      // vertex color flag
        bool        isFakeEnvironment;  // fake environment flag
        this() { }
        this(ModelMaterial mat) 
        { 
            copy(mat); 
            name = mat.name;
            isVertexColor = mat.isVertexColor;
            isFakeEnvironment = mat.isFakeEnvironment;
        }
    };


    // object data class
    static class ModelObject {
    public:
        char[]          name;                   // object name
        vl.vector3[]    vertex;                 // vertex data array
        FACE[]          face;                   // face data array
        bool            isSmoothing = false;    // smoothing flag
        float           angleFacet = 0;         // facet angle for smooth shading [degree]
        ubyte           flagMirror = 0;         // mirroring 5bit flag (3,4bit for type,0,1,2bit for axis flag)
        bool            isLocking = false;      // locking flag
        bool            isVisible = false;      // visible flag
        uint            color = 0xffffffff;     // color
        vl.vector3      vcMinimum;              // minimum vertex exsiting range (calculate after loading)
        vl.vector3      vcMaximum;              // maximum vertex exsiting range (calculate after loading)

    public:
        // constructor
        this() { }
        
        this(ModelObject obj)
        {
            name = obj.name;
            vertex.length = obj.vertex.length;
            vertex[] = obj.vertex;
            face.length = obj.face.length;
            face[] = obj.face;
            isSmoothing = obj.isSmoothing;
            angleFacet = obj.angleFacet;
            flagMirror = obj.flagMirror;
            isLocking = obj.isLocking;
            isVisible = obj.isVisible;
            color = obj.color;
            vcMinimum = obj.vcMinimum;
            vcMaximum = obj.vcMaximum;
        }
        
        // access to vertex pointer
        vl.vector3*  opIndex(int iFace, int iVert)
        {
            return &vertex[face[iFace].idx[iVert]];
        }

    public:
        // operate(vertex0, vertex1, vertex2, vertex_normal0, vertex_normal1, vertex_normal2, face_normal)
        ModelObject operate(void delegate(vl.vector3*, vl.vector3*, vl.vector3*, vl.vector3*, vl.vector3*, vl.vector3*, vl.vector3*) dg) 
        {
            for (int i=0; i<face.length; i++) {
                dg(&vertex[face[i].idx[0]],  &vertex[face[i].idx[1]],  &vertex[face[i].idx[2]],
                   &face[i].normalVertex[0], &face[i].normalVertex[1], &face[i].normalVertex[2], 
                   &face[i].normalFace);
            }
            return this;
        }

        // multiply vl.matrix and vertex
        ModelObject mul_vertex(vl.matrix4* m)
        {
            for (int i=0; i<vertex.length; i++) vl.mul(&vertex[i], m);
            return this;
        }
        
        // multiply vl.matrix and normal vl.vector
        ModelObject mul_normal(vl.matrix4* m)
        {
            if (isSmoothing) {
                for (int i=0; i<face.length; i++) {
                    vl.mul_w0(&face[i].normalFace, m);
                    vl.mul_w0(&face[i].normalVertex[0], m);
                    vl.mul_w0(&face[i].normalVertex[1], m);
                    vl.mul_w0(&face[i].normalVertex[2], m);
                }
            } else {
                for (int i=0; i<face.length; i++) {
                    vl.mul_w0(&face[i].normalFace, m);
                    face[i].normalVertex[] = face[i].normalFace;
                }
            }
            return this;
        }

        // calculate normal vector
        ModelObject calcNormal()
        {
            int i,j,k,idx;
            
            // calculate face normal vector
            for (i=0; i<face.length; i++) {
                vl.calcNormal(&face[i].normalFace, this[i,0], this[i,1], this[i,2]);
                face[i].normalVertex[] = face[i].normalFace;
            }
            
            // calculate vertex normal vector
            if (isSmoothing) {
                scope ushort[3][] sum_count = new ushort[3][face.length];   // averaging counter
                float facetcos = cos(degree2radian(angleFacet));            // cos of facet angle
                
                // search and sum up shared vertex's normal vector
                for (i=0; i<face.length-1; i++) {
                    for (j=0; j<3; ++j) {
                        for (k=i+1; k<face.length; ++k) {
                            idx = searchFace(k, face[i].idx[j]);    // search vertex shared face
                            if (k == -1) break;                     // finish searching

                            // sum up vertex normal when angle of 2 faces less than facet angle.
                            if (vl.dot(&face[k].normalFace, &face[i].normalFace) > facetcos) {
                                vl.add(&face[i].normalVertex[j]   ,&face[k].normalFace);
                                vl.add(&face[k].normalVertex[idx] ,&face[i].normalFace);
                                ++sum_count[i][j];
                                ++sum_count[k][idx];
                            }
                        }
                    }
                }
                
                // averaging
                for (i=0; i<face.length-1; i++) {
                    vl.scale(&face[i].normalVertex[0], 1.0f/cast(float)(sum_count[i][0]+1));
                    vl.scale(&face[i].normalVertex[1], 1.0f/cast(float)(sum_count[i][1]+1));
                    vl.scale(&face[i].normalVertex[2], 1.0f/cast(float)(sum_count[i][2]+1));
                }
            }
            
            return this;
        }

        // calculate vertex existing range
        void calcRange()
        {
            if (vertex.length == 0) return;
            vl.copy(&vcMinimum, &vertex[0]);
            vl.copy(&vcMaximum, &vertex[0]);
            for (int i=1; i<vertex.length; i++) {
                vl.copyMin(&vcMaximum, &vcMaximum, &vertex[i]);
                vl.copyMax(&vcMinimum, &vcMinimum, &vertex[i]);
            }
        }

        // search face that has specifyed vertex
        // (take start index by idx, and give result. return value is vertex index.)
        int searchFace(inout int idx, int v)
        {
            for (; idx<face.length; idx++) {
                if (face[idx].idx[0] == v) return 0;
                if (face[idx].idx[1] == v) return 1;
                if (face[idx].idx[2] == v) return 2;
            }
            // search finished
            idx = -1;
            return -1;
        }

        // search face that has specifyed vertexes, (return) and (return+1) is shared vertex indexes.
        int searchFace(inout int idx, int v0, int v1)
        {
            for (; idx<face.length; idx++) {
                if ((face[idx].idx[0]==v0 && face[idx].idx[1]==v1) ||
                    (face[idx].idx[0]==v1 && face[idx].idx[1]==v0))   return 0;
                if ((face[idx].idx[1]==v0 && face[idx].idx[2]==v1) ||
                    (face[idx].idx[1]==v1 && face[idx].idx[2]==v0))   return 1;
                if ((face[idx].idx[2]==v0 && face[idx].idx[0]==v1) ||
                    (face[idx].idx[2]==v1 && face[idx].idx[0]==v0))   return 2;
            }
            // search finished
            idx = -1;
            return -1;
        }
        
        // make CArray
        void makeArray(CArray array, bool isSmooth)
        {
            int i, j;
            int start_idx = array.getDataCount();
            array.expand(face.length * 3);

            // copy vertex
            for (i=0, j=start_idx; i<face.length; i++) {
                vl.copy(array.getVertex(j++), this[i, 0]);
                vl.copy(array.getVertex(j++), this[i, 1]);
                vl.copy(array.getVertex(j++), this[i, 2]);
            }

            // copy normal
            if (array.isSupportNormal()) {
                for (i=0, j=start_idx; i<face.length; i++) {
                    vl.copy(array.getNormal(j++), &face[i].normalVertex[0]);
                    vl.copy(array.getNormal(j++), &face[i].normalVertex[1]);
                    vl.copy(array.getNormal(j++), &face[i].normalVertex[2]);
                }
            }

            // copy vertex color
            if (array.isSupportVertexColor()) {
                int jmax = array.getDataCount();
                for (j=start_idx; j<jmax; ++j) {
                    *(cast(uint*)array.getColor(j)) = color;
                }
            }

            // copy texture coordination
            if (array.isSupportTexCoord()) {
                for (i=0, j=start_idx; i<face.length; i++) {
                    vl.copy(array.getTexCoord(j++), &face[i].uv[0]);
                    vl.copy(array.getTexCoord(j++), &face[i].uv[1]);
                    vl.copy(array.getTexCoord(j++), &face[i].uv[2]);
                }
            }
        }
        
        // make wire frame CArray
        void makeWireFrame(CArray array, float fFacetDegree)
        {
            int i, j, k, add, linked_vert;

            scope ushort[]  wfIndex = new ushort[face.length * 6];
            scope bool[3][] wfDone  = new bool[3][face.length];

            float facetcos = cos(degree2radian(180-fFacetDegree));
            
            // create wire
            for (i=0, add=0; i<face.length; i++) {
                for (j=0; j<3; j++) {
                    k = (j+1) % 3;
                    if (!wfDone[i][j]) {
                        int idx = i+1;
                        // search vertex shared face
                        linked_vert = searchFace(idx, face[i].idx[j], face[i].idx[k]);
                        if (idx == -1) {
                            // make wire when no faces linked.
                            wfIndex[add++] = face[i].idx[j];
                            wfIndex[add++] = face[i].idx[k];
                        } else {
                            wfDone[idx][linked_vert] = true;
                            if (vl.dot(&face[i].normalFace, &face[idx].normalFace) <= facetcos) {
                                // make wire when angle of 2 faces less than facet angle.
                                wfIndex[add++] = face[i].idx[j];
                                wfIndex[add++] = face[i].idx[k];
                            }
                        }
                    }
                }
            }

            // copy wire lines to array
            array.expand(add);
            for (i=0; i<add; i++) { vl.copy(array.getVertex(i), &vertex[wfIndex[i]]); }
        }
    };




// member value
//--------------------------------------------------
public:
    ModelMaterial[]     material;           // material data
    ModelObject[]       object;             // object data

    GLuint              uidDisplayList;     // display list ID
    int                 countDisplayList;   // display list count
    vl.vector3          vcMinimum;          // minimum vertex exsiting range
    vl.vector3          vcMaximum;          // maximum vertex exsiting range

    char[]              nameDir;            // directory name

    CTextureManager     managerTexture;     // texture manager
    
    // default texture manager
    static CTextureManager defalutTextureManager;
    static CTextureManager getDefaultTextureManager() { return defalutTextureManager; }
    static this() { defalutTextureManager = new CTextureManager(); }




// constructor
//--------------------------------------------------
public:
    this()
    {
        managerTexture = defalutTextureManager;
    }
    
    
    this(CSolidModel mdl)
    {
        copy(mdl);
    }
    
    
    this(CTextureManager mng)
    {
        managerTexture = mng;
    }
    
    
    this(char[] nameFile, bool reg_texture=true, bool cal_normal=true)
    {
        this();
        if (!load(nameFile, reg_texture, cal_normal)) free();
    }
    
    
    ~this()
    {
        free(false);
    }




// memory opearation
//--------------------------------------------------
protected:
    // allocate new materials
    void _allocMaterial(int cnt)
    {
        material = new ModelMaterial[cnt];
        foreach(inout mat; material) mat = new ModelMaterial;
    }


    // new object
    ModelObject _newObject()
    {
        ModelObject newObject = new ModelObject;
        object ~= newObject;
        return newObject;
    }


    // free
    void free(bool deleteList=true)
    {
        if (deleteList) deleteDisplayList();
        foreach(obj; object)   { delete(obj); }
        foreach(mat; material) { delete(mat); }
        object.length = 0;
        material.length = 0;
    }


    // copy
    CSolidModel copy(CSolidModel src)
    {
        uidDisplayList = 0;
        countDisplayList = 0;
        vcMinimum = src.vcMinimum;
        vcMaximum = src.vcMaximum;
        nameDir = src.nameDir;
        managerTexture = src.managerTexture;
        
        material.length = src.material.length;
        for (int i=0; i<src.material.length; i++) {
            material[i] = new ModelMaterial(src.material[i]);
        }

        object.length = src.object.length;
        for (int i=0; i<src.object.length; i++) {
            object[i] = new ModelObject(src.object[i]);
        }
        
        return this;
    }




// file opearation
//--------------------------------------------------
public:
    bool load(char[] nameFile, bool reg_texture=true, bool cal_normal=true)
    {
        // file open
        char[][] line;
        try { line = splitlines(cast(char[])std.file.read(nameFile)); }
        catch (Exception ex) { return false; }
        nameDir = getDirName(nameFile);
        if (nameDir.length != 0) nameDir ~= '\\';

        // check header
        if (!_checkHeader(line)) return false;

        // read body
        if (!_readBody(line)) return false;

        // calculate normal
        if (cal_normal) calcNormal();
        
        // register texture
        if (reg_texture) registerTexture();

        return true;
    }


protected:
    // read 1 line, returns token list
    bool _nextLine(inout char[][] line)
    {
        while (line.length > 1) {
            line = line[1..$];
            if (line[0].length == 0) continue;  // ignore void line
            if (line[0][0] == '#')   continue;  // ignore comment
            return true;
        }
        return false;
    }


    // skip to next chunk
    bool _nextChunk(inout char[][] line)
    {
        while (_nextLine(line)) { 
            if (line[0] == "}") return _nextLine(line);
        }
        return false;
    }


    // check header
    bool _checkHeader(inout char[][] line)
    {
        if (line.length < 2 ||
            line[0] != "Metasequoia Document" || 
            line[1] != "Format Text Ver 1.0") return false;
        line=line[1..$];
        return _nextLine(line);
    }


    // read body
    bool _readBody(inout char[][] line)
    {
        while (true) {
            scope char[][] tok = std.string.split(line[0]);
            if (tok.length == 1 && tok[0] == "Eof") return true;
            if (tok.length <  2 || tok[length-1] != "{") return false;
            switch(tok[0]) {
            case "Material":    // material chunk
                if (!_readMaterialChunk(line, atoi(tok[1]))) return false;
                break;
            case "Object":      // object chunk
                if (!_readObjectChunk(line, _newObject())) return false;
                break;
            default:            // unknown chunk
                if (!_nextChunk(line)) return false;
            }
        }
        assert(false);
    }
    
    
    // read materal chunk
    bool _readMaterialChunk(inout char[][] line, int material_count)
    {
        // allocate material
        if (material_count == 0) return false;
        _allocMaterial(material_count);

        // read
        for (int idx=0; idx<material_count; idx++) {
            if (!_nextLine(line)) return false;
            if (!_readMaterialLine(line[0], material[idx])) return false;
        }
        
        // go to next chunk
        return _nextChunk(line);
    }


    // read material line
    bool _readMaterialLine(char[] line, ModelMaterial mat)
    {
        // get name
        int s_idx = std.string.find(line, '"');
        if (s_idx == -1) return false;
        int e_idx = std.string.find(line[s_idx+1..$-1], '"');
        if (e_idx == -1) return false;
        mat.name = line[s_idx+1..s_idx+e_idx+1].dup;

        // get information
        float[9]   param = [0.5f, 0.5f, 0.5f, 1.0f, 0.3f, 1.0f, 0.0f, 0.5f, 0.0f];
        char[][3]  filename;
        
        foreach (m; RegExp(r"([a-z_]*)\((.*)\)").search(line)) {
            if (m.match(2)==null) return false;
            switch(m.match(1)) {
            case "col":     foreach(i, str; std.string.split(m.match(2))) { param[i] = atof(str); } break;
            case "vcol":    mat.isVertexColor = cast(bool)atoi(m.match(2)); break;
            case "amb":     param[4] = atof(m.match(2)); break;
            case "dif":     param[5] = atof(m.match(2)); break;
            case "spc":     param[6] = atof(m.match(2)); break;
            case "power":   param[7] = atof(m.match(2)); break;
            case "emi":     param[8] = atof(m.match(2)); break;
            case "tex":     filename[0] = nameDir ~ m.match(2)[1..$-1]; break;
            case "aplane":  filename[1] = nameDir ~ m.match(2)[1..$-1]; break;
            case "bump":    filename[2] = nameDir ~ m.match(2)[1..$-1]; break;
            case "shader":
            case "proj_type":   
            case "proj_pos":
            case "proj_scale":
            case "proj_angle":
                break;
            default:
                break;
            }
        }

        // set material parameter
        mat.setColor(param[0], param[1], param[2], param[3]);
        mat.setLight(param[4], param[5], param[6], param[7], param[8]);

        // load texture
        if (filename[0].length != 0) {
            if (filename[1].length == 0) mat.setTexture(managerTexture.load(filename[0], 3));           // normal texture
            else                         mat.setTexture(managerTexture.load(filename[0], filename[1])); // with alpha map
        }
        
        // if bump map is filled, support fake environment map
        mat.isFakeEnvironment = (filename[2].length != 0);

        return true;
    }


    // read object chunk
    bool _readObjectChunk(inout char[][] line, ModelObject obj)
    {
        // get name
        int s_idx = std.string.find(line[0], '"');
        if (s_idx == -1) return false;
        int e_idx = std.string.find(line[0][s_idx+1..$-1], '"');
        if (e_idx == -1) return false;
        obj.name = line[0][s_idx+1..s_idx+e_idx+1].dup;
        
        // read
        float[3] param = [0.5f, 0.5f, 0.5f];
        while (_nextLine(line)) {
            char[][] tok = std.string.split(line[0]);  // get token
            if (tok.length == 0) continue;  // void line
            switch (tok[0]) {
            case "color":       if (tok.length>=4) { for (int i=0; i<3; i++) { param[i] = atof(tok[i+1]); } } break;
            case "shading":     if (tok.length>=2) { obj.isSmoothing = cast(bool)atoi(tok[1]); } break;
            case "facet":       if (tok.length>=2) { obj.angleFacet = atof(tok[1]); } break;
            case "mirror":      if (tok.length>=2) { obj.flagMirror |= atoi(tok[1]) << 3; } break;
            case "mirror_axis": if (tok.length>=2) { obj.flagMirror |= atoi(tok[1]); } break;
            case "visible":     if (tok.length>=2) { obj.isVisible = cast(bool)atoi(tok[1]); } break;
            case "locking":     if (tok.length>=2) { obj.isLocking = cast(bool)atoi(tok[1]); } break;
            case "vertex":      if (tok.length>=3) { if (!_readVertexData(line, obj, atoi(tok[1]))) return false; } break;
            case "face":        if (tok.length>=3) { if (!_readFaceData  (line, obj, atoi(tok[1]))) return false; } break;
            case "}":           _nextLine(line); return true;   // success to read all data
            default:            break;                          // skip unknown tag
            }
        }
        
        return false;
    }


    // read vertex data
    bool _readVertexData(inout char[][] line, ModelObject obj, int size)
    { 
        // allocate
        obj.vertex.length =size;

        int idx=0;
        while (_nextLine(line)) {
            char[][] tok = std.string.split(line[0]);
            if (tok.length == 3) {
                if (idx == obj.vertex.length) return false;
                obj.vertex[idx].x = atof(tok[0]);
                obj.vertex[idx].y = atof(tok[1]);
                obj.vertex[idx].z = atof(tok[2]);
                idx++;
                continue;
            }
            if (tok[0] == "}") {
                return (idx == obj.vertex.length);
            }
            break;
        }
        return false;
    }


    // read face data
    bool _readFaceData(inout char[][] line, ModelObject obj, int size)
    { 
        // tempolary area
        scope byte[]          vcnt = new byte[size];        // vertex count
        scope short[]         mat  = new short[size];       // material index
        scope ushort[4][]     idx  = new ushort[4][size];   // vertex index
        scope vl.vector2[4][] uv   = new vl.vector2[4][size];  // texture coordination
        int i, j;
        
        // read data
        int fcount = 0;                          // face count
        for (i=0; i<size; i++) {
            if (!_nextLine(line)) return false;
            if (!_readFaceLine(line[0], vcnt[i], mat[i], idx[i], uv[i])) return false;
            fcount += vcnt[i] - 2;
        }
        if (!_nextLine(line))      return false;
        if (strip(line[0]) != "}") return false;    // check end

        // allocate face
        obj.face = new FACE[fcount];

        // copy face data, divide quadrangle to 2 triangles
        for (i=0, j=0; i<size; i++, j++) {
            obj.face[j].mat   = mat[i];
            obj.face[j].idx[] = idx[i][0..3];
            obj.face[j].uv[]  = uv[i][0..3];
            
            if (vcnt[i] == 4) {
                ++j;
                obj.face[j].mat       = mat[i];
                obj.face[j].idx[0]    = idx[i][0];
                obj.face[j].idx[1..3] = idx[i][2..4];
                obj.face[j].uv[0]     = uv[i][0];
                obj.face[j].uv[1..3]  = uv[i][2..4];
            }
        }
        
        return true;
    }


    // read face line
    bool _readFaceLine(char[] line, out byte vcount, out short mat, ushort[] index, vl.vector2[] uv)
    {
        // vertex count
        vcount = atoi(line);
        if (vcount != 3 && vcount != 4) { return false; }

        char[][] tok;
        // analyse token
        foreach (m; RegExp(r"(V|M|UV)\((.*)\)").search(line)) {
            if (m.match(2)==null) return false;
            switch(m.match(1)) {
            case "V":
                tok = std.string.split(m.match(2));
                if (vcount != tok.length) return false;
                for (int i=0; i<vcount; i++) { index[i] = atoi(tok[i]); }
                break;
            case "M":
                mat = atoi(m.match(2));
                break;
            case "UV":
                tok = std.string.split(m.match(2));
                if (vcount*2 != tok.length) return false;
                for (int i=0; i<vcount; i++) {
                    uv[i].x = atof(tok[i*2]);
                    uv[i].y = 1-atof(tok[i*2+1]);
                }
                break;
            default:
                return false;
            }
        }
        
        return true;
    }




// calculation
//--------------------------------------------------
public:
    // multiply matrix
    CSolidModel mul(vl.matrix4* m)
    {
        foreach(obj; object) obj.mul_vertex(m).calcNormal();
        return this;
    }
    
    CSolidModel mul(int idx, vl.matrix4* m)
    {
        object[idx].mul_vertex(m).calcNormal();
        return this;
    }

    // multiply quarternion
    CSolidModel mul(vl.vector4* q)
    { 
        vl.matrix4 m; 
        vl.loadRotate(&m, q); 
        foreach(obj; object) obj.mul_vertex(&m).mul_normal(&m); 
        return this; 
    }
    
    CSolidModel mul(int idx, vl.vector4* q)
    {
        vl.matrix4 m;
        vl.loadRotate(&m, q);
        object[idx].mul_vertex(&m).mul_normal(&m);
        return this;
    }

    // translate vertex
    CSolidModel translate(float x, float y, float z)
    {
        for(int i=0; i<object.length; i++) translate(i,x,y,z); 
        return this;
    }
    
    CSolidModel translate(int idx, float x, float y, float z)
    {
        for(int i=0; i<object[idx].vertex.length; i++) vl.add(&object[idx].vertex[i], x, y, z);
        return this;
    }

    // scale vertex
    CSolidModel scale(float x, float y, float z)
    {
        for(int i=0; i<object.length; i++) scale(i,x,y,z); 
        return this;
    }
    
    CSolidModel scale(int idx, float x, float y, float z)
    {
        for(int i=0; i<object[idx].vertex.length; i++) vl.scale(&object[idx].vertex[i], x, y, z);
        return this;
    }

    // rotate vertex
    CSolidModel rotateX(float deg) 
    {
        vl.vector2 rot;
        vl.setAngle(&rot, degree2radian(deg));
        foreach(obj; object) {
            for(int i=0; i<obj.vertex.length; i++) vl.mulX(&obj.vertex[i], &rot);
        }
        return this;
    }
    
    CSolidModel rotateX(int idx, float deg) {
        vl.vector2 rot;
        vl.setAngle(&rot, degree2radian(deg));
        for(int i=0; i<object[idx].vertex.length; i++) vl.mulX(&object[idx].vertex[i], &rot);
        return this;
    }
    
    // rotate vertex
    CSolidModel rotateY(float deg)
    {
        vl.vector2 rot;
        vl.setAngle(&rot, degree2radian(deg));
        foreach(obj; object) {
            for(int i=0; i<obj.vertex.length; i++) vl.mulY(&obj.vertex[i], &rot);
        }
        return this;
    }
    
    CSolidModel rotateY(int idx, float deg) {
        vl.vector2 rot;
        vl.setAngle(&rot, degree2radian(deg));
        for(int i=0; i<object[idx].vertex.length; i++) vl.mulY(&object[idx].vertex[i], &rot);
        return this;
    }
    
    // rotate vertex
    CSolidModel rotateZ(float deg)
    {
        vl.vector2 rot;
        vl.setAngle(&rot, degree2radian(deg));
        foreach(obj; object) {
            for(int i=0; i<obj.vertex.length; i++) vl.mulZ(&obj.vertex[i], &rot);
        }
        return this;
    }
    
    CSolidModel rotateZ(int idx, float deg)
    {
        vl.vector2 rot;
        vl.setAngle(&rot, degree2radian(deg));
        for(int i=0; i<object[idx].vertex.length; i++) vl.mulZ(&object[idx].vertex[i], &rot);
        return this;
    }


    // calculate face/vertex normal vector
    CSolidModel calcNormal()
    {
        foreach(obj; object) obj.calcNormal();
        return this;
    }
    
    CSolidModel calcNormal(int idx)
    {
        object[idx].calcNormal();
        return this;
    }


    // caluclate vertex existing range
    CSolidModel calcRange() { 
        foreach(obj; object) { 
            obj.calcRange();
            vl.copyMin(&vcMaximum, &vcMaximum, &obj.vcMaximum);
            vl.copyMax(&vcMinimum, &vcMinimum, &obj.vcMinimum);
        } 
        return this; 
    }
    
    CSolidModel calcRange(int idx) {
        object[idx].calcRange();
        vl.copyMin(&vcMaximum, &vcMaximum, &object[idx].vcMaximum);
        vl.copyMax(&vcMinimum, &vcMinimum, &object[idx].vcMinimum);
        return this;
    }
    

    // set material to fake environment map
    void setFakeEnvironmentMaterial(int iMaterialIndex, CTexture texEnvmap)
    {
        if (iMaterialIndex < material.length) {
            material[iMaterialIndex].isFakeEnvironment = true;
            if (texEnvmap !is null) { material[iMaterialIndex].texture = texEnvmap; }
        }
    }




// drawing
//--------------------------------------------------
public:
    // draw object
    void draw(bool isSmooth=false) { for (int i=0; i<object.length; i++) { draw(i, isSmooth); } }
    void draw(int idx, bool isSmooth=false)
    {
        ModelObject obj = object[idx];
        DRAWSUB sub_idx;
        DRAWSUB vnormal = (isSmooth && obj.isSmoothing) ? DRAWSUB.VERTEX_NORMAL : DRAWSUB.NOFLAG;

        for (int i=0; i<obj.face.length;) {
            sub_idx = _drawChangeMaterial(obj.face[i].mat) | vnormal;   // change material
            glBegin(GL_TRIANGLES);                                      // drawing sub routine
            _funcDrawSub[sub_idx](obj, i);
            glEnd();
        }
    }


    // draw object ignoring materials
    void drawSolid(bool isSmooth=false) { for (int i=0; i<object.length; i++) { drawSolid(i, isSmooth); } }
    void drawSolid(int idx, bool isSmooth=false)
    {
        ModelObject obj  = object[idx];
        FACE[]      face = obj.face;

        glBegin(GL_TRIANGLES);
        if (isSmooth && obj.isSmoothing) {
            for (int i=0; i<face.length; i++) {
                glNormal3fv(&face[i].normalVertex[2].x);
                glVertex3fv(&obj[i,2].x);
                glNormal3fv(&face[i].normalVertex[1].x);
                glVertex3fv(&obj[i,1].x);
                glNormal3fv(&face[i].normalVertex[0].x);
                glVertex3fv(&obj[i,0].x);
            }
        } else {
            for (int i=0; i<face.length; i++) {
                glNormal3fv(&face[i].normalFace.x);
                glVertex3fv(&obj[i,2].x);
                glVertex3fv(&obj[i,1].x);
                glVertex3fv(&obj[i,0].x);
            }
        }
        glEnd();
    }

    // resiter all textures
    void registerTexture() { for (int i=0; i<material.length; i++) { material[i].register(); } }

    // unresiter all textures
    void unregisterTexture() { for (int i=0; i<material.length; i++) { material[i].unregister(); } }


protected:
    // drawing sub-routin flag
    enum DRAWSUB {
        NOFLAG        = 0x00,
        VERTEX_NORMAL = 0x01,
        TEXTURE_UV    = 0x02,
        ENVIRONMENT   = 0x05,   // include DRAWSUB_VERTEX_NORMAL
        max           = 6
    };
    
    // drawing sub-routin
    static void function(ModelObject, inout int)[DRAWSUB.max] _funcDrawSub = [
        &_drawSub, &_drawSubS, &_drawSubT, &_drawSubST, &_drawSubEnv, &_drawSubEnv
    ];

    // change material (sub routine for draw())
    DRAWSUB _drawChangeMaterial(int mat)
    {
        if (mat == -1) return DRAWSUB.NOFLAG;
        material[mat].draw();
        if (material[mat].isFakeEnvironment) return DRAWSUB.ENVIRONMENT;
        if (material[mat].texture is null)   return DRAWSUB.NOFLAG;
        return DRAWSUB.TEXTURE_UV;
    }

    // draw with nothing
    static void _drawSub(ModelObject obj, inout int i)
    {
        FACE[] face = obj.face;
        int    mat  = face[i].mat;
        for (; i<face.length; i++) {
            if (face[i].mat != mat) return; 
            glNormal3fv(&face[i].normalFace.x);
            glVertex3fv(&obj[i,2].x);
            glVertex3fv(&obj[i,1].x);
            glVertex3fv(&obj[i,0].x);
        }
    }

    // draw with vertex-normal
    static void _drawSubS(ModelObject obj, inout int i)
    {
        FACE[] face = obj.face;
        int    mat  = face[i].mat;
        for (; i<face.length; i++) {
            if (face[i].mat != mat) return; 
            glNormal3fv(&face[i].normalVertex[2].x);
            glVertex3fv(&obj[i,2].x);
            glNormal3fv(&face[i].normalVertex[1].x);
            glVertex3fv(&obj[i,1].x);
            glNormal3fv(&face[i].normalVertex[0].x);
            glVertex3fv(&obj[i,0].x);
        }
    }

    // draw with texture
    static void _drawSubT(ModelObject obj, inout int i)
    {
        FACE[] face = obj.face;
        int    mat  = face[i].mat;
        for (; i<face.length; i++) {
            if (face[i].mat != mat) return; 
            glNormal3fv(&face[i].normalFace.x);
            glTexCoord2fv(&face[i].uv[2].x);
            glVertex3fv(&obj[i,2].x);
            glTexCoord2fv(&face[i].uv[1].x);
            glVertex3fv(&obj[i,1].x);
            glTexCoord2fv(&face[i].uv[0].x);
            glVertex3fv(&obj[i,0].x);
        }
    }

    // draw with vertex-normal and texture
    static void _drawSubST(ModelObject obj, inout int i)
    {
        FACE[] face = obj.face;
        int    mat  = face[i].mat;
        for (; i<face.length; i++) {
            if (face[i].mat != mat) return; 
            glNormal3fv(&face[i].normalVertex[2].x);
            glTexCoord2fv(&face[i].uv[2].x);
            glVertex3fv(&obj[i,2].x);
            glNormal3fv(&face[i].normalVertex[1].x);
            glTexCoord2fv(&face[i].uv[1].x);
            glVertex3fv(&obj[i,1].x);
            glNormal3fv(&face[i].normalVertex[0].x);
            glTexCoord2fv(&face[i].uv[0].x);
            glVertex3fv(&obj[i,0].x);
        }
    }

    // draw vertex-normal as a texture uv
    static void _drawSubEnv(ModelObject obj, inout int i)
    {
        FACE[] face = obj.face;
        int    mat  = face[i].mat;

        glPushAttrib(GL_ENABLE_BIT);

        glEnable(GL_TEXTURE_2D);
        glDisable(GL_LIGHTING);
        for (; i<face.length; i++) {
            if (face[i].mat != mat) return; 
            glTexCoord2f(face[i].normalVertex[2].x+0.5f, face[i].normalVertex[2].y+0.5f);
            glVertex3fv(&obj[i,2].x);
            glTexCoord2f(face[i].normalVertex[1].x+0.5f, face[i].normalVertex[1].y+0.5f);
            glVertex3fv(&obj[i,1].x);
            glTexCoord2f(face[i].normalVertex[0].x+0.5f, face[i].normalVertex[0].y+0.5f);
            glVertex3fv(&obj[i,0].x);
        }

        glPopAttrib();
    }




// compose array
//--------------------------------------------------
public:
    // make CArray
    void makeArray(CArray array, bool isSmooth=false)
    {
        foreach(obj; object) obj.makeArray(array, isSmooth);
    }
    
    void makeArray(int idx, CArray array, bool isSmooth=false)
    {
        array.free();
        object[idx].makeArray(array, isSmooth);
    }


    // make wire frame CArray
    void makeWireFrame(CArray array, float fFacetDegree=180.0f)
    {
        foreach(obj; object) obj.makeWireFrame(array, fFacetDegree);
    }
    
    void makeWireFrame(int idx, CArray array, float fFacetDegree=180.0f)
    {
        array.free();
        object[idx].makeWireFrame(array, fFacetDegree);
    }




// operate
//--------------------------------------------------
public:
    // operate ModelObject
    CSolidModel operate(int delegate(ModelObject) dg)
    {
        foreach(obj; object) { if (dg(obj)!=0) break; }
        return this;
    }


    // operate vertex
    CSolidModel operate(int delegate(vl.vector3*) dg)
    {
        foreach(obj; object) {
            for (int i=0; i<obj.vertex.length; i++) {
                if (dg(&obj.vertex[i])!=0) return this;
            }
        }
        return this;
    }




// reference
//--------------------------------------------------
public:
    // is empty ?
    bool isEmpty()
    {
        return (object.length == 0);
    }
    

    // find object by name
    int  findObjectIndex(char[] name)
    { 
        for(int i=0; i<object.length; i++) { 
            if (object[i].name == name) return i; 
        }
        return -1;
    }
    
        
    // find material by name
    int  findMaterialIndex(char[] name)
    { 
        for(int i=0; i<material.length; i++) { 
            if (material[i].name == name) return i; 
        }
        return -1;
    }




// display list
//--------------------------------------------------
public:
    // create display list (isDivided = false; make divided display list for each object)
    CSolidModel compile(bool isDivided=false, bool isSmooth=false)
    {
        registerTexture();
        if (isDivided) {
            uidDisplayList = glGenLists(object.length);
            if (uidDisplayList == 0) throw new Exception("Failed to create display list");
            countDisplayList = object.length;
            for (int i=0; i<countDisplayList; i++) {
                glNewList(uidDisplayList+i, GL_COMPILE);
                    draw(i, isSmooth);
                glEndList();
            }
        } else {
            uidDisplayList = glGenLists(1);
            if (uidDisplayList == 0) throw new Exception("Failed to create display list");
            countDisplayList = 1;
            glNewList(uidDisplayList, GL_COMPILE);
                draw(isSmooth);
            glEndList();
        }
        return this;
    }


    // delete display list
    CSolidModel deleteDisplayList()
    {
        if (countDisplayList <= 0) return this;
        glDeleteLists(uidDisplayList, countDisplayList);
        countDisplayList = 0;
        uidDisplayList = 0;
        return this;
    }
    
    
    // call display list
    void callList()
    {
        for (int i=0; i<countDisplayList; i++) { glCallList(uidDisplayList + i); }
    }
    

    // call display list with object index
    void callList(int idx)
    {
        if (idx < countDisplayList) glCallList(uidDisplayList + idx);
    }




// static cast operation
//--------------------------------------------------
public:
    // make display list from file
    static int opCast(char[] nameFile, bool isSmooth=false)
    {
        scope CSolidModel mdl = new CSolidModel(nameFile);
        if (mdl.isEmpty()) return -1;
        mdl.compile(false, isSmooth);
        int uid = mdl.uidDisplayList;
        return uid;
    }


    // make array from file
    static CArray opCast(CArray array, char[] nameFile, bool isSmooth=false)
    {
        scope CSolidModel mdl = new CSolidModel(nameFile);
        if (!mdl.isEmpty()) mdl.makeArray(array, isSmooth);
        return array;
    }


    // make wire frame from file
    static CArray opCast(CArray wf, char[] nameFile, float facet=180.0f)
    {
        scope CSolidModel mdl = new CSolidModel(nameFile);
        if (!mdl.isEmpty()) mdl.makeWireFrame(wf, facet);
        return wf;
    }


    // make array and wire frame from file
    static CArray opCast(CArray array, CArray wf, char[] nameFile, bool isSmooth=false, float facet=180.0f)
    {
        scope CSolidModel mdl = new CSolidModel(nameFile);
        if (!mdl.isEmpty()) {
            mdl.makeArray(array, isSmooth);
            mdl.makeWireFrame(wf, facet);
        }
        return wf;
    }
};




