//------------------------------------------------------------------------------------------
// OpenGL Texture Class
//------------------------------------------------------------------------------------------
module util_opengl.texture;


private import opengl;
private import openglu;
private import std.stdio;
private import util.list;
private import util.math;
private import util.color;
private import util.bitmap;




class CTexture
{
// member value
//---------------------------------------------
public:
    ubyte[] data = null;            // pixel data
    int     width = 0;              // width
    int     height = 0;             // height
    int     log2width = 0;          // log2(width)
    int     sizebytePixel = 1;      // pixel data size (GL_LUMINANS=1, GL_RGB=3, GL_RGBA=4)
    char[]  nameFile;               // file name
    char[]  nameFileAlpha;          // file name for alpha map

    uint    uidName = 0;            // texture name for openGL (0=un-registerd)

    bool    isMipmapping = true;                        // mip-mpapping texture ?
    int     flagExpansion = GL_NEAREST_MIPMAP_NEAREST;  // expantion filter flag
    int     flagReduction = GL_NEAREST;                 // reduction filter flag
    int     flagWrapU = GL_REPEAT;                      // u-axis wrapping filter flag 
    int     flagWrapV = GL_REPEAT;                      // v-axis wrapping filter flag 

    GLenum  format = GL_LUMINANCE;  // pixel data format (GL_LUMINANCE | GL_RGB | GL_RGBA) (auto-selected with uidTexture) 
    GLenum  target = GL_TEXTURE_2D; // texture target (GL_TEXTURE_1D | GL_TEXTURE_2D)




// constructor
//---------------------------------------------
public:
    this() 
    {
    }
    
    
    this(int width_, int height_, int sizebytePixel_, ubyte[] pix=null) 
    {
        create(width_, height_, sizebytePixel_, pix);
    }
    
    
    this(char[] nameFile_, int sizebytePixel_) 
    {
        load(nameFile_, sizebytePixel_);
    }
    
    
    this(char[] nameFile_, char[] nameFileAlpha_)
    {
        load(nameFile_, nameFileAlpha_);
    }

    
    this(uint col, char[] nameFileAlpha_)
    {
        load(col, nameFileAlpha_);
    }

    
    ~this()
    {
        destroy();
    }




// initialize
//---------------------------------------------
public:
    // create texture
    CTexture create(int width_, int height_, int sizebytePixel_, ubyte[] pix=null)
    {
        _alloc(width_, height_, sizebytePixel_);
        if (pix == null) data[] = 0;
        else data[0..pix.length] = pix[];
        return this;
    }
    
    
    // load file by specifyed pixel size
    CTexture load(char[] nameFile_, int sizebytePixel_)
    {
        int bmpbit, w, h;
        scope ubyte[] pxl;
        scope ubyte[] tbl;
        
        // read file
        readBitmap(nameFile_, w, h, bmpbit, pxl, tbl);
        
        // set bit data
        if (bmpbit!=24 && bmpbit!=8 && bmpbit!=1) throw new Exception(nameFile_ ~ "; format is not supported");

        _alloc(w, h, sizebytePixel_);
        switch(sizebytePixel) {
        case 1: convertBitmap8 (data, pxl, tbl, bmpbit); break;
        case 3: convertBitmap24(data, pxl, tbl, bmpbit, false); break;
        case 4: convertBitmap32(data, pxl, tbl, bmpbit, false); break;
        }
        
        return this;
    }
    
    
    // load file as alpha map file
    CTexture load(uint col, char[] nameFileAlpha_)
    {
        int bmpbit, w, h;
        scope ubyte[] pxl;
        scope ubyte[] pxl_alpha;
        scope ubyte[] tbl;
        
        // read alpha map file
        readBitmap(nameFileAlpha_, w, h, bmpbit, pxl, tbl);

        // set alpha map
        if (bmpbit!=24 && bmpbit!=8 && bmpbit!=1) throw new Exception(nameFileAlpha_ ~ "; format is not supported");

        // create and set color
        create(w, h, 4)[] = col;

        convertBitmap8(pxl_alpha, pxl, tbl, bmpbit);
        setAlphaMap(data, pxl_alpha);
        
        return this;
    }
    
    
    // load file with alpha map file
    CTexture load(char[] nameFile_, char[] nameFileAlpha_)
    {
        int bmpbit, w, h;
        scope ubyte[] pxl;
        scope ubyte[] pxl_alpha;
        scope ubyte[] tbl;
        
        // load color map file
        load(nameFile_, 4);
        
        // read alpha map file
        readBitmap(nameFileAlpha_, w, h, bmpbit, pxl, tbl);
        // set alpha map
        if (bmpbit!=24 && bmpbit!=8 && bmpbit!=1) throw new Exception(nameFileAlpha_ ~ "; format is not supported");
        convertBitmap8(pxl_alpha, pxl, tbl, bmpbit);
        setAlphaMap(data, pxl_alpha);
        
        return this;
    }
    
    
    // load file with transparent color
    CTexture load_trans(char[] nameFile_, uint colTrans)
    {
        // load color map file and set transparent color
        load(nameFile_, 4);
        return transparent(colTrans);
    }
    
    
    // set transparent color
    CTexture transparent(uint colTrans)
    {
        transparentAlphaColor(data, colTrans);
        return this;
    }
    
    
    // free bit data
    void destroy(bool withUnregister=true)
    {
        if (withUnregister) unregister();
        _free();
    }




// ooperation
//---------------------------------------------
public:
    // setFormat() must be called before register()
    CTexture setFormat(bool isMipmapping_, bool isReductionLinear, bool isExpansionLinear, bool isURepeating, bool isVRepeating)
    {
        isMipmapping = isMipmapping_;
        if (isMipmapping)  flagReduction = (isReductionLinear) ? (GL_LINEAR_MIPMAP_NEAREST) : (GL_NEAREST_MIPMAP_NEAREST);
        else               flagReduction = (isReductionLinear) ? (GL_LINEAR) : (GL_NEAREST);
        flagExpansion = (isExpansionLinear) ? (GL_LINEAR) : (GL_NEAREST);
        flagWrapU = (isURepeating) ? (GL_REPEAT) : (GL_CLAMP);
        flagWrapV = (isVRepeating) ? (GL_REPEAT) : (GL_CLAMP);

        if (uidName != 0) { unregister(); register(); }
        
        return this;
    }
    
    
    // chengeFormat() can be called after register()
    CTexture changeFormat(bool isReductionLinear, bool isExpansionLinear, bool isURepeating, bool isVRepeating)
    {
        if (isMipmapping)  flagReduction = (isReductionLinear) ? (GL_LINEAR_MIPMAP_NEAREST) : (GL_NEAREST_MIPMAP_NEAREST);
        else               flagReduction = (isReductionLinear) ? (GL_LINEAR) : (GL_NEAREST);
        flagExpansion = (isExpansionLinear) ? (GL_LINEAR) : (GL_NEAREST);
        flagWrapU = (isURepeating) ? (GL_REPEAT) : (GL_CLAMP);
        flagWrapV = (isVRepeating) ? (GL_REPEAT) : (GL_CLAMP);

        if (uidName == 0) register();
        
        bind();
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, flagReduction);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, flagExpansion);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, flagWrapU);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, flagWrapV);
        unbind();

        return this;
    }
    
    
    // translate bit data to graphic memory
    CTexture register()
    {
        if (uidName == 0) glGenTextures(1, &uidName);

        bind();
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, flagReduction);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, flagExpansion);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, flagWrapU);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, flagWrapV);
        if (target == GL_TEXTURE_1D) {
            if (isMipmapping) gluBuild1DMipmaps(GL_TEXTURE_1D,    sizebytePixel, width,    format, GL_UNSIGNED_BYTE, cast(void*)data);
            else              glTexImage1D     (GL_TEXTURE_1D, 0, sizebytePixel, width, 0, format, GL_UNSIGNED_BYTE, cast(void*)data);
        } else {
            if (isMipmapping) gluBuild2DMipmaps(GL_TEXTURE_2D,    sizebytePixel, width, height,    format, GL_UNSIGNED_BYTE, cast(void*)data);
            else              glTexImage2D     (GL_TEXTURE_2D, 0, sizebytePixel, width, height, 0, format, GL_UNSIGNED_BYTE, cast(void*)data);
        }
        unbind();
        
        return this;
    }
    
    
    // free bit data on graphic memory
    void unregister()
    {
        if (uidName == 0) return;
        glDeleteTextures(1, &uidName);
        uidName = 0;
    }
    
    
    // translate bit data to graphic memory after register()
    void update()
    {
        if (uidName == 0) return;
        bind();
        if (target == GL_TEXTURE_1D) {
            if (isMipmapping) gluBuild1DMipmaps(GL_TEXTURE_1D, sizebytePixel, width, format, GL_UNSIGNED_BYTE, cast(void*)data);
            else              glTexSubImage1D  (GL_TEXTURE_1D, 0, 0,          width, format, GL_UNSIGNED_BYTE, cast(void*)data);
        } else {
            if (isMipmapping) gluBuild2DMipmaps(GL_TEXTURE_2D, sizebytePixel, width, height, format, GL_UNSIGNED_BYTE, cast(void*)data);
            else              glTexSubImage2D  (GL_TEXTURE_2D, 0, 0, 0,       width, height, format, GL_UNSIGNED_BYTE, cast(void*)data);
        }
        unbind();
    }
    
    
    // glBindTexture()
    void bind()
    {
        glBindTexture(target, uidName);
    }
    
    
    // glBindTexture(0)
    void unbind()
    {
        glBindTexture(target, 0);
    }

    
    // glCopyTexImage2D()
    void copyTexImage2D(int x, int y)
    {
        if (uidName == 0) return;
        bind();
        glCopyTexImage2D(target, 0, format, x, y, width, height, 0);
        unbind();
    }

    
    // glDrawPixels()
    void drawPixels(float x, float y)
    {
        glRasterPos2f(x, y);
        glDrawPixels(width, height, format, GL_UNSIGNED_BYTE, cast(void*)data);
    }

    
    // glReadPixels()
    void readPixels(int x, int y)
    {
        glPixelStorei(GL_PACK_ALIGNMENT ,4);
        glReadPixels(x, y, width, height, GL_RGBA, GL_UNSIGNED_BYTE, cast(void*)data);
    }




// system memory operation
//---------------------------------------------
protected:
    // allocate bit data area in system memory
    void _alloc(int width_, int height_, int sizebytePixel_)
    {
        switch (sizebytePixel_) {
        case 1:  format = GL_LUMINANCE;       break;
        case 2:  format = GL_LUMINANCE_ALPHA; break;
        case 3:  format = GL_RGB;             break;
        case 4:  format = GL_RGBA;            break;
        default: format = GL_RGBA; sizebytePixel_=4; break;
        }

        width  = calc2N(width_);
        height = calc2N(height_);
        log2width = calcLog2(width);
        sizebytePixel = sizebytePixel_;
        
        data.length = width * height * sizebytePixel;
    }


    // free bit data area in system memory
    void _free()
    {
        format = GL_LUMINANCE;
        width = 0;
        height = 0;
        log2width = 0;
        sizebytePixel = 1;
        nameFile = "";
        nameFileAlpha = "";

        delete(data);
        data = null;
    }




// reference
//--------------------------------------------------
public:
    uint opIndex(int x, int y)
    {
        int idx = ((y << log2width) + x) * sizebytePixel;
        switch(sizebytePixel) {
        case 3:  return rgba2color(data[idx..idx+3]);
        case 4:  return rgba2color(data[idx..idx+4]);
        default: return data[idx];
        }
    }


    ubyte[] opSlice()
    {
        return data;
    }
    
    
    ubyte[] opSliceAssign(uint col)
    {
        int i;
        ubyte[4] ub;
        color2rgba(ub, col);
        switch(sizebytePixel) {
        case 3:  for(i=0; i<data.length; i+=3) { data[i]=ub[0]; data[i+1]=ub[1]; data[i+2]=ub[2]; } break;
        case 4:  for(i=0; i<data.length; i+=4) { data[i]=ub[0]; data[i+1]=ub[1]; data[i+2]=ub[2]; data[i+3]=ub[3]; } break;
        default: data[] = cast(ubyte)col; break;
        }
        return data;
    }


    bool isEmpty() { return (data == null); }
    bool isLoaded() { return (nameFile.length != 0); }
    bool isRegistered() { return (uidName != 0); }
};




//--------------------------------------------------------------------------------
// CTextureManager; same file open once.
//--------------------------------------------------------------------------------
class CTextureManager
{
private:
    list!(CTexture) listTexture;


public:
    // constructor
    this()
    {
        listTexture = new list!(CTexture);
    }

    
    // register to list
    CTexture append(CTexture tex)
    {
        return listTexture.push_front(tex);
    }

    
    // remove from list
    CTexture remove(CTexture tex)
    {
        foreach(list!(CTexture).iterator it; listTexture) { 
            if (it() is tex) { listTexture.remove(it); break; }
        }
        return tex;
    }
    
    
    // new
    CTexture create() 
    {
        return append(new CTexture());
    }

    
    // create
    CTexture create(int width, int height, int sizebytePixel)
    {
        return append(new CTexture(width, height, sizebytePixel));
    }
    
    
    // load 
    CTexture load(char[] nameFile, int sizebytePixel)
    {
        CTexture tex = find(nameFile);
        if (tex !is null) return tex;
        tex = new CTexture();
        if (!tex.load(nameFile, sizebytePixel)) { delete(tex); return null; }
        return append(tex);
    }

    
    // load with alpha map
    CTexture load(char[] nameFile, char[] nameFileAlpha)
    {
        CTexture tex = find(nameFile, nameFileAlpha);
        if (tex !is null) return tex;
        tex = new CTexture();
        if (!tex.load(nameFile, nameFileAlpha)) { delete(tex); return null; }
        return append(tex);
    }

    
    // load with transparent color
    CTexture load_trans(char[] nameFile, uint colTrans)
    {
        CTexture tex = find(nameFile);
        if (tex !is null) return tex;
        tex = new CTexture();
        if (!tex.load_trans(nameFile, colTrans)) { delete(tex); return null; }
        return append(tex);
    }
    
    
    // find file
    CTexture find(char[] nameFile)
    {
        foreach(CTexture tex; listTexture) { 
            if (tex.nameFile==nameFile && tex.nameFileAlpha.length==0) return tex;
        }
        return null;
    }
    
    
    // find file
    CTexture find(char[] nameFile, char[] nameFileAlpha)
    {
        foreach(CTexture tex; listTexture) { 
            if (tex.nameFile==nameFile && tex.nameFileAlpha==nameFileAlpha) return tex;
        }
        return null;
    }

    
    // call all register()
    int registerAll()
    {
        int cnt = 0;
        foreach(CTexture tex; listTexture) { if (tex.register()) cnt++; }
        return cnt;
    }

    
    // call all unregister()
    void unregisterAll()
    {
        foreach(CTexture tex; listTexture) { tex.unregister(); }
    }
    
    
    // foreach operation
    int opApply(int delegate(inout CTexture) dg)
    {
        foreach(CTexture tex; listTexture) { 
            int res=dg(tex);
            if (res!=0) return res;
        }
        return 0;
    }
};






