//--------------------------------------------------------------------------------
//  Control class
//--------------------------------------------------------------------------------
module util_sdl.controler;

private import sdl;
private import util_sdl.joystick;




//--------------------------------------------------------------------------------
//  Control key enums
//--------------------------------------------------------------------------------
public enum PAD {
    LEFT    = 0,
    RIGHT   = 1,
    UP      = 2,
    DOWN    = 3,
    BUTTON0 = 4,
    BUTTON1 = 5,
    BUTTON2 = 6,
    BUTTON3 = 7,
    BUTTON4 = 8,
    BUTTON5 = 9,
    BUTTON6 = 10,
    BUTTON7 = 11,
    BUTTON8 = 12,
    BUTTON9 = 13,
    BUTTON10= 14,
    ESCAPE  = 15,
    max
}


const int JS_BUTTON_MAX = 12;




//--------------------------------------------------------------------------------
//  Controler class
//--------------------------------------------------------------------------------
class CControler
{
private:
    uint                m_id;                   // index of control
    SDLKey[][PAD.max]   m_tablePADtoKB;         // Look up table from PAD to keyboard
    PAD[JS_BUTTON_MAX]  m_tableJStoPAD;         // Look up table from Joystick button to PAD

    ushort              m_flagInput;            // 16 bit pad input status
    uint[PAD.max]       m_counterKeyFrame;      // Key pressing frame counter
    uint                m_rangeJoystickFree;    // Joy stick free range
    
    CSDLJoystick        m_joystick;             // Instance of joystick




public:
    this()
    {
        reset();
        m_joystick = new CSDLJoystick;
    }
    
    
    void reset()
    {
        int i;
        
        // Default Joystick Setting
        for (i=0; i<JS_BUTTON_MAX; i++) {
            assignJoystick(i, cast(PAD)(PAD.BUTTON0 + i));
        }
        setJoystickFreeRange(0.4f);

        // Reset flag and counter
        m_flagInput = 0;
        for (i=0; i<PAD.max; i++) {
            m_counterKeyFrame[i] = 0;
        }
    }
    


// Joy stick operation
//--------------------------------------------------
public:
    // Open joystick
    bool openJoystick()
    {
        return m_joystick.open(m_id);
    }


    // Close joystick
    void closeJoystick()
    {
        m_joystick.close();
    }
    
    
    // Joystick opened ?
    bool isJoystickOpened()
    {
        return m_joystick.isOpened();
    }



// assign
//--------------------------------------------------
public:
    // assign keys
    void assign(SDLKey sdlkey, PAD key)
    {
        if (key < PAD.max) m_tablePADtoKB[key] ~= sdlkey;
    }
    
    
    // clear all assgin key table
    void clearAssign()
    {
        foreach (SDLKey[] kt; m_tablePADtoKB) { kt.length = 0; }
    }
    
    
    // assign joy stick button
    void assignJoystick(uint jsb_index, PAD key)
    {
        m_tableJStoPAD[jsb_index] = key;
    }
    

    // set joy stick free range
    void setJoystickFreeRange(float range)
    {
        m_rangeJoystickFree = cast(int)(32765.0f * range);
    }




public:
    // Is pressed ?
    bool isPressed(PAD key)
    {
        return (m_counterKeyFrame[key] != 0);
    }
    
    
    // Is start pressing from this frame ?
    bool isHitted(PAD key)
    {
        return (m_counterKeyFrame[key] == 1);
    }
    
    
    // Get pressing frame counter.
    int getFrameCounter(PAD key)
    {
        return m_counterKeyFrame[key];
    }
    

    // Get x and y input
    void getXY(inout int x, inout int y)
    {
        x = (m_counterKeyFrame[PAD.RIGHT] != 0) - (m_counterKeyFrame[PAD.LEFT] != 0);
        y = (m_counterKeyFrame[PAD.UP]    != 0) - (m_counterKeyFrame[PAD.DOWN] != 0);
    }
    
    
    // Get 16bit flags of pad status
    ushort getFlags()
    {
        return m_flagInput;
    }




// (internal functions)
public:
    void _internal_set_id(uint id)
    {
        m_id = id;
    }
    
    
    void _internal_update_flag(ubyte* pKeyTable)
    {
        int  i;
        bool isPressed[PAD.max];
        
        // check keyboard
        for (i=0; i<PAD.max; i++) {
            isPressed[i] = false;
            foreach (SDLKey key; m_tablePADtoKB[i]) {
                isPressed[i] |= (pKeyTable[key] == SDL_PRESSED);
            }
        }
        
        // check joystick
        if (m_joystick.isOpened()) {
            i = m_joystick.getXAxis();
            isPressed[PAD.LEFT]  |= (i < -m_rangeJoystickFree);
            isPressed[PAD.RIGHT] |= (i >  m_rangeJoystickFree);
            i = m_joystick.getYAxis();
            isPressed[PAD.UP]    |= (i < -m_rangeJoystickFree);
            isPressed[PAD.DOWN]  |= (i >  m_rangeJoystickFree);
            for (i=0; i<JS_BUTTON_MAX; i++) {
                isPressed[i] |= m_joystick.isPressed(m_tableJStoPAD[i]);
            }
        }
        
        // calculate flag
        m_flagInput = 0;
        for (i=0; i<PAD.max; i++) {
            m_flagInput |= (isPressed[i] << i);
        }
    }


    void _internal_update_counter()
    {
        int  i;
        for (i=0; i<PAD.max; i++) {
            if (m_flagInput & (1 << i)) m_counterKeyFrame[i]++;
            else                        m_counterKeyFrame[i] = 0;
        }
    }
}

